from collections import OrderedDict

from staramr.results.AMRDetectionSummary import AMRDetectionSummary

"""
Summarizes both ResFinder and PointFinder database results into a single table.
"""


class AMRDetectionSummaryResistance(AMRDetectionSummary):

    def __init__(self, files, resfinder_dataframe, quality_module_dataframe,pointfinder_dataframe=None, plasmidfinder_dataframe=None, mlst_dataframe=None):
        """
        Creates a new AMRDetectionSummaryResistance.
        :param files: The list of genome files we have scanned against.
        :param resfinder_dataframe: The pd.DataFrame containing the ResFinder results.
        :param quality_module_dataframe: The pd.DataFrame containing the genome size, N50 value, number of contigs under our user defined minimum length
        as well as the results of our quality metrics (pass or fail) and the corresponding feedback
        :param pointfinder_dataframe: The pd.DataFrame containing the PointFinder results.
        :param plasmidfinder_dataframe: The pd.DataFrame containing the PlasmidFinder results.
        """
        super().__init__(files, resfinder_dataframe,quality_module_dataframe, pointfinder_dataframe, plasmidfinder_dataframe, mlst_dataframe)

    def _aggregate_phenotype(self, phenotype_series):
        flattened_phenotype_list = [y.strip() for x in list(phenotype_series) for y in
                                    x.split(self.SEPARATOR)]

        # Only remove None if there is more than one entry in this list
        if len(flattened_phenotype_list) > 1 and 'None' in flattened_phenotype_list:
            flattened_phenotype_list.remove('None')

        uniq_phenotype = OrderedDict.fromkeys(flattened_phenotype_list)

        return (self.SEPARATOR + ' ').join(list(uniq_phenotype))

    def _aggregate_gene(self, gene_series):
        return (self.SEPARATOR + ' ').join(list(gene_series))

    def _compile_results(self, df):
        df_summary = df.copy()

        # Used to sort by gene names, ignoring case
        df_summary['Gene.Lower'] = df['Gene'].str.lower()

        # Compiles the gene/phenotype results into a single entry per isolate (groupby)
        df_summary = df_summary \
            .sort_values(by=['Gene.Lower']) \
            .groupby(['Isolate ID'], sort=True) \
            .aggregate({'Gene': self._aggregate_gene,
                'Predicted Phenotype': self._aggregate_phenotype})
        return df_summary[['Gene', 'Predicted Phenotype']]

    def _get_detailed_negative_columns(self):
        return ['Isolate ID', 'Gene', 'Predicted Phenotype', 'Start', 'End']

    def _get_summary_empty_values(self):
        return {'Genotype': 'None', 'Predicted Phenotype': 'Sensitive'}

    def _get_summary_resistance_columns(self):
        return ['Genotype', 'Predicted Phenotype', 'Plasmid']

    def _get_detailed_summary_columns(self):
        return ['Gene', 'Data Type', 'Predicted Phenotype', '%Identity', '%Overlap', 'HSP Length/Total Length', 'Contig', 'Start',
                'End', 'Accession']

    def _include_phenotype(self):
        return True



#!/usr/bin/python3
import shutil
import os
import sys
import tempfile
import subprocess

# TODO:
# Make script independent of current working directory
# Make script able to store indexed files in a directory not named
# 'kma_indexing'

# This scripts installs the PointFinder database for using KMA
# KMA should be installed before running this script
# The scripts assumes that it is placed together with the ResFinder species
# directories
#
# First clone the repository:
# git clone https://bitbucket.org/genomicepidemiology/resfinder_db.git

# Check if executable kma_index program is installed, if not promt the user for
# path


# Function for easy error printing
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


# KMA version
#KMA_VERSION = "latest"


interactive = True
if len(sys.argv) >= 2:
    kma_index = sys.argv[1]
    if "non_interactive" in sys.argv:
        interactive = False
else:
    kma_index = "kma_index"

print(str(sys.argv))

while shutil.which(kma_index) is None:
    eprint("KMA index program, {}, does not exist or is not executable"
           .format(kma_index))
    ans = None
    if(interactive):
        ans = input("Please input path to executable kma_index program or"
                    "choose one of the options below:\n"
                    "\t1. Install KMA using make, index db, then remove KMA.\n"
                    "\t2. Exit\n")

    if(ans == "2" or ans == "q" or ans == "quit" or ans == "exit"):
        eprint("Exiting!\n\n"
               "Please install executable KMA programs in order to install"
               "this database.\n\n"
               "KMA can be obtained from bitbucked:\n\n"
               "git clone"
               "https://bitbucket.org/genomicepidemiology/kma.git"
               )
        sys.exit()

    if(ans == "1" or ans is None):
        if(shutil.which("git") is None):
            sys.exit("Attempt to automatically install KMA failed.\n"
                     "git does not exist or is not executable.")
        org_dir = os.getcwd()

        # Create temporary directory
        tempdir = tempfile.TemporaryDirectory()
        os.chdir(tempdir.name)

        try:
            subprocess.run(
                ["git", "clone",
                 "https://bitbucket.org/genomicepidemiology/kma.git"],
                check=True)
            os.chdir("kma")
        except subprocess.CalledProcessError:
            eprint("Installation in temporary directory with make failed "
                   "at the git cloning step")
            os.chdir(org_dir)

        try:
            subprocess.run(["make"])
        except subprocess.CalledProcessError:
            eprint("Installation in temporary directory with make failed "
                   "at the make step.")
            os.chdir(org_dir)

        kma_index = "{}/kma/kma_index".format(tempdir.name)
        os.chdir(org_dir)
        if shutil.which(kma_index) is None:
            eprint("Installation in temporary directory with make failed "
                   "at the test step.")
            os.chdir(org_dir)
            kma_index = "kma_index"
            if(not interactive):
                ans = "2"

    if(ans is not None and ans != "1" and ans != "2"):
        kma_index = ans
        if shutil.which(kma_index) is None:
            eprint("Path, {}, is not an executable path. Please provide "
                   "absolute path\n".format(ans))


# Index databases

# Use config_file to go through database dirs
config_file = open("config", "r")
for line in config_file:
    if line.startswith("#"):
        continue
    else:
        line = line.rstrip().split("\t")
        drug = line[0].strip()
        # for each dir index the fasta files
        os.system("{0} -i {1}.fsa -o ./{1}".format(kma_index, drug))
os.system("{0} -i *.fsa -o ./all".format(kma_index))

config_file.close()

eprint("Done")



from staramr.blast.results.pointfinder.PointfinderHitHSP import PointfinderHitHSP
from staramr.blast.results.pointfinder.nucleotide.NucleotideMutationPosition import NucleotideMutationPosition


class PointfinderHitHSPRNA(PointfinderHitHSP):

    def __init__(self, file, blast_record):
        """
        Creates a new PointfinderHitHSPRNA.
        :param file: The input file.
        :param blast_record: The Bio.Blast.Record this hit came from.
        """
        super().__init__(file, blast_record)

    def _get_mutation_positions(self, start):
        mutation_positions = []

        amr_seq = self.get_amr_gene_seq()
        genome_seq = self.get_genome_contig_hsp_seq()

        amr_pos = 0

        for i in range(len(amr_seq)):

            # Insertion: "-" in the reference:
            if amr_seq[i] == "-":
                # left side
                offset = i - amr_pos  # accounting for string index and reference index possibly being different
                mutation = NucleotideMutationPosition(i - 1, amr_seq, genome_seq, start, offset=offset)
                mutation_positions.append(mutation)
            # Mismatch or Deletion:
            elif (amr_seq[i] != genome_seq[i]):
                offset = i - amr_pos
                mutation = NucleotideMutationPosition(i, amr_seq, genome_seq, start, offset=offset)
                mutation_positions.append(mutation)
                amr_pos += 1
            # Match:
            else:
                amr_pos += 1

        return mutation_positions



import copy
import logging
import os
import pandas as pd
from os import path
import re
from collections import Counter
from typing import List, Dict, Optional

from Bio import SeqIO
from pandas import DataFrame

from staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase import PlasmidfinderBlastDatabase
from staramr.blast.pointfinder.PointfinderBlastDatabase import PointfinderBlastDatabase
from staramr.blast.resfinder.ResfinderBlastDatabase import ResfinderBlastDatabase
from staramr.blast.results.BlastResultsParser import BlastResultsParser
from staramr.blast.results.plasmidfinder.BlastResultsParserPlasmidfinder import BlastResultsParserPlasmidfinder
from staramr.blast.results.pointfinder.BlastResultsParserPointfinder import BlastResultsParserPointfinder
from staramr.blast.results.resfinder.BlastResultsParserResfinder import BlastResultsParserResfinder
from staramr.results.AMRDetectionSummary import AMRDetectionSummary
from staramr.results.QualityModule import QualityModule

logger = logging.getLogger("AMRDetection")

"""
A Class to handle scanning files for AMR genes.
"""


class AMRDetection:

    def __init__(self, resfinder_database: ResfinderBlastDatabase, amr_detection_handler,
                 pointfinder_database: PointfinderBlastDatabase = None,
                 include_negative_results: bool = False, output_dir: str = None, genes_to_exclude: list = [],
                 plasmidfinder_database: PlasmidfinderBlastDatabase = None) -> None:
        """
        Builds a new AMRDetection object.
        :param resfinder_database: The staramr.blast.resfinder.ResfinderBlastDatabase for the particular ResFinder database.
        :param amr_detection_handler: The staramr.blast.JobHandler to use for scheduling BLAST jobs.
        :param pointfinder_database: The staramr.blast.pointfinder.PointfinderBlastDatabase to use for the particular PointFinder database.
        :param plasmidfinder_database: The staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase for the particular PlasmidFinder database.
        :param include_negative_results:  If True, include files lacking AMR genes in the resulting summary table.
        :param output_dir: The directory where output fasta files are to be written into (None for no output fasta files).
        :param genes_to_exclude: A list of gene IDs to exclude from the results.
        """
        self._resfinder_database = resfinder_database
        self._amr_detection_handler = amr_detection_handler
        self._pointfinder_database = pointfinder_database
        self._plasmidfinder_database = plasmidfinder_database
        self._include_negative_results = include_negative_results

        if pointfinder_database is None:
            self._has_pointfinder = False
        else:
            self._has_pointfinder = True

        self._output_dir = output_dir

        self._genes_to_exclude = genes_to_exclude

    def _create_amr_summary(self, files: List[str], resfinder_dataframe: DataFrame,quality_module_dataframe: DataFrame,
                            pointfinder_dataframe: Optional[BlastResultsParserPointfinder],
                            plasmidfinder_dataframe: DataFrame, mlst_dataframe: DataFrame) -> DataFrame:
        amr_detection_summary = AMRDetectionSummary(files, resfinder_dataframe,quality_module_dataframe,
                                                    pointfinder_dataframe, plasmidfinder_dataframe, mlst_dataframe)
        return amr_detection_summary.create_summary(self._include_negative_results)

    def _create_detailed_amr_summary(self, files: List[str], resfinder_dataframe: DataFrame,quality_module_dataframe: DataFrame,
                                     pointfinder_dataframe: Optional[BlastResultsParserPointfinder],
                                     plasmidfinder_dataframe: DataFrame, mlst_dataframe: DataFrame) -> DataFrame:
        amr_detection_summary = AMRDetectionSummary(files, resfinder_dataframe,quality_module_dataframe,
                                                    pointfinder_dataframe, plasmidfinder_dataframe, mlst_dataframe)
        return amr_detection_summary.create_detailed_summary(self._include_negative_results)

    def _create_resfinder_dataframe(self, resfinder_blast_map: Dict, pid_threshold: float, plength_threshold: int,
                                    report_all: bool) -> DataFrame:
        resfinder_parser = BlastResultsParserResfinder(resfinder_blast_map, self._resfinder_database, pid_threshold,
                                                       plength_threshold, report_all, output_dir=self._output_dir,
                                                       genes_to_exclude=self._genes_to_exclude)
        return resfinder_parser.parse_results()

    def _create_pointfinder_dataframe(self, pointfinder_blast_map: Dict, pid_threshold: float, plength_threshold: int,
                                      report_all: bool) -> DataFrame:
        pointfinder_parser = BlastResultsParserPointfinder(pointfinder_blast_map, self._pointfinder_database,
                                                           pid_threshold, plength_threshold, report_all,
                                                           output_dir=self._output_dir,
                                                           genes_to_exclude=self._genes_to_exclude)
        return pointfinder_parser.parse_results()

    def _create_plasmidfinder_dataframe(self, plasmidfinder_blast_map: Dict[str, BlastResultsParser],
                                        pid_threshold: float, plength_threshold: int,
                                        report_all: bool) -> DataFrame:
        plasmidfinder_parser = BlastResultsParserPlasmidfinder(plasmidfinder_blast_map, self._plasmidfinder_database,
                                                               pid_threshold,
                                                               plength_threshold, report_all,
                                                               output_dir=self._output_dir,
                                                               genes_to_exclude=self._genes_to_exclude)
        return plasmidfinder_parser.parse_results()

    def create_quality_module_dataframe(self,files,genome_size_lower_bound,genome_size_upper_bound,minimum_N50_value,
                                        minimum_contig_length,unacceptable_num_contigs) ->DataFrame:
        quality_module = QualityModule(files,genome_size_lower_bound,genome_size_upper_bound,minimum_N50_value,
                                        minimum_contig_length,unacceptable_num_contigs)

        return quality_module.create_quality_module_dataframe()

    def _generate_empty_columns(self, row: list, max_cols: int, cur_cols: int) -> list:
        if(cur_cols < max_cols):
            for i in range(max_cols-cur_cols):
                row.append('-')

        return row

    def _create_mlst_dataframe(self, mlst_data: str) -> DataFrame:

        columns = ['Isolate ID', 'Scheme', 'Sequence Type']
        curr_data = []
        max_columns = 0
        extension = None

        mlst_split = mlst_data.splitlines()

        # Parse and format the current row
        for row in mlst_split:
            array_format = re.split('\t', row)
            num_columns = len(array_format)

            # We want the file name without the extension
            array_format[0] = path.basename(path.splitext(array_format[0])[0])

            if max_columns < num_columns:
                max_columns = num_columns

            curr_data.append(array_format)

        # Go through each row and append additional columns for the dataframes
        curr_data = list(map(lambda x: self._generate_empty_columns(x, max_columns, len(x)), curr_data))

        # Append Locus Column names if any
        locus_columns = max_columns - len(columns)
        if locus_columns > 0:
            for x in range(0, locus_columns):
                columns.append(("Locus {}").format(x+1))

        mlst_dataframe = pd.DataFrame(curr_data, columns=columns)
        mlst_dataframe = mlst_dataframe.set_index('Isolate ID')

        return mlst_dataframe

    def run_amr_detection(self,files, pid_threshold, plength_threshold_resfinder, plength_threshold_pointfinder,
                          plength_threshold_plasmidfinder, genome_size_lower_bound,genome_size_upper_bound,
                          minimum_N50_value,minimum_contig_length,unacceptable_num_contigs,
                          report_all=False, ignore_invalid_files=False, mlst_scheme=None) -> None:
        """
        Scans the passed files for AMR genes.
        :param files: The files to scan.
        :param pid_threshold: The percent identity threshold for BLAST results.
        :param plength_threshold_resfinder: The percent length overlap for BLAST results (resfinder).
        :param plength_threshold_pointfinder: The percent length overlap for BLAST results (pointfinder).
        :param plength_threshold_plasmidfinder: The percent length overlap for BLAST results (plasmidfinder).
        :param genome_size_lower_bound: The lower bound for the genome size as defined by the user for quality metrics
        :param genome_size_upper_bound: The upper bound for the genome size as defined by the user for quality metrics
        :param minimum_N50_value: The minimum N50 value as defined by the user for quality metrics
        :param minimum_contig_length: The minimum contig length as defined by the user for quality metrics
        :param unacceptable_num_contigs: The number of contigs in a file, equal to or above our minimum contig length, for which to raise a flag as defined by the user for quality metrics
        :param report_all: Whether or not to report all blast hits.
        :param ignore_invalid_files: Skips the invalid input files if set.
        :param mlst_scheme: Specifys scheme name MLST uses if set.
        :return: None
        """

        files_copy = copy.deepcopy(files)
        files = self._validate_files(files_copy, ignore_invalid_files)

        self._quality_module_dataframe=self.create_quality_module_dataframe(files,genome_size_lower_bound,genome_size_upper_bound,minimum_N50_value,minimum_contig_length,unacceptable_num_contigs)

        self._amr_detection_handler.run_blasts_mlst(files, mlst_scheme)

        resfinder_blast_map = self._amr_detection_handler.get_resfinder_outputs()
        self._resfinder_dataframe = self._create_resfinder_dataframe(resfinder_blast_map, pid_threshold,
                                                                     plength_threshold_resfinder, report_all)

        plasmidfinder_blast_map = self._amr_detection_handler.get_plasmidfinder_outputs()
        self._plasmidfinder_dataframe = self._create_plasmidfinder_dataframe(plasmidfinder_blast_map, pid_threshold,
                                                                             plength_threshold_plasmidfinder,
                                                                             report_all)

        mlst_data = self._amr_detection_handler.get_mlst_outputs()
        self._mlst_dataframe = self._create_mlst_dataframe(mlst_data)

        self._pointfinder_dataframe = None
        if self._has_pointfinder:
            pointfinder_blast_map = self._amr_detection_handler.get_pointfinder_outputs()
            self._pointfinder_dataframe = self._create_pointfinder_dataframe(pointfinder_blast_map, pid_threshold,
                                                                             plength_threshold_pointfinder, report_all)

        self._summary_dataframe = self._create_amr_summary(files, self._resfinder_dataframe,self._quality_module_dataframe,
                                                           self._pointfinder_dataframe, self._plasmidfinder_dataframe, self._mlst_dataframe)

        self._detailed_summary_dataframe = self._create_detailed_amr_summary(files, self._resfinder_dataframe,self._quality_module_dataframe,
                                                                             self._pointfinder_dataframe,
                                                                             self._plasmidfinder_dataframe,
                                                                             self._mlst_dataframe)

    def _validate_files(self, files: List[str], ignore_invalid_files: bool) -> List[str]:
        total_files = len(files)
        removeable_files = []

        for file in files:
            # Check if the file is not a directory
            if os.path.isdir(file):
                if ignore_invalid_files:
                    logger.warning('--ignore-invalid-files is set, skipping directory {}'.format(file))
                    removeable_files.append(file)
                else:
                    raise Exception(
                        'Directory {} is invalid, please use --ignore-invalid-files to skip over this directory'.format(
                            file))
            else:
                # Will raise an error if the input returns an empty generator on non-FASTA files, returns a boolean
                validInput = any(SeqIO.parse(file, "fasta"))

                if not validInput:
                    if ignore_invalid_files:
                        logger.warning('--ignore-invalid-files is set, skipping file {}'.format(file))
                        removeable_files.append(file)
                    else:
                        raise Exception(
                            'File {} is invalid, please use --ignore-invalid-files to skip over invalid input files'.format(
                                file))
                else:
                    # Check if there are any duplicate sequence id's in the valid files
                    record = []
                    # Store all the sequence id's in a list
                    for sequence in SeqIO.parse(file, "fasta"):
                        record.append(sequence.id)

                    duplicates = []

                    # Each sequence contains a tuple (sequence id, frequency)
                    for sequence in (Counter(record)).items():
                        if sequence[1] > 1:
                            # We want the sequence id's that are duplicates
                            duplicates.append(sequence[0])

                    # Raise an error if there's any duplicates in the file
                    if len(duplicates) > 0:
                        raise Exception(
                            'File {} contains the following duplicate sequence IDs: {}'.format(file, duplicates))

        # Check to see if the invalid file is not the only file in the directory
        if total_files == len(removeable_files):
            raise Exception('Cannot produce output due to no valid input files')

        # Remove the skipped files
        if ignore_invalid_files:
            for file in removeable_files:
                files.remove(file)

        return files

    def get_mlst_results(self):
        """
        Gets a pd.DataFrame for the MLST results.
        :return: A pd.DataFrame for the MLST results.
        """

        return self._mlst_dataframe

    def get_resfinder_results(self):
        """
        Gets a pd.DataFrame for the ResFinder results.
        :return: A pd.DataFrame for the ResFinder results.
        """
        return self._resfinder_dataframe

    def get_pointfinder_results(self):
        """
        Gets a pd.DataFrame for the PointFinder results.
        :return: A pd.DataFrame for the PointFinder results.
        """
        return self._pointfinder_dataframe

    def get_plasmidfinder_results(self):
        """
        Gets a pd.DataFrame for the PlasmidFinder results.
        :return: A pd.DataFrame for the PlasmidFinder results.
        """
        self._plasmidfinder_dataframe = self._plasmidfinder_dataframe.rename({'Gene':'Plasmid'}, axis=1)
        return self._plasmidfinder_dataframe

    def get_summary_results(self):
        """
        Gets a pd.DataFrame for a summary table of the results.
        :return: A pd.DataFrame for a summary table of the results.
        """
        return self._summary_dataframe

    def get_detailed_summary_results(self):
        """
        Gets a pd.DataFrame for a detailed summary table of the results.
        :return: A pd.DataFrame for a detailed summary table of the results.
        """

        self._detailed_summary_dataframe = self._detailed_summary_dataframe.rename({'Gene':'Data'}, axis=1)
        return self._detailed_summary_dataframe



import logging
from os import path
from typing import Set

import pandas as pd
from pandas import DataFrame
from Bio import SeqIO

logger = logging.getLogger("QualityModule")

"""
Summarizes the quality metric feedback into a single table.
"""


class QualityModule:

    def __init__(self, files, genome_size_lower_bound,genome_size_upper_bound,minimum_N50_value,minimum_contig_length,unacceptable_num_contigs) -> None:
        """
        Constructs an object for summarizing our quality module.
        :param files: The list of genome files we have scanned against.
        :param genome_size_lower_bound: The lower bound for the genome size as defined by the user for quality metrics
        :param genome_size_upper_bound: The upper bound for the genome size as defined by the user for quality metrics
        :param minimum_N50_value: The minimum N50 value as defined by the user for quality metrics
        :param minimum_contig_length: The minimum contig length as defined by the user for quality metrics
        :param unacceptable_num_contigs: The number of contigs in a file, equal to or above our minimum contig length, for which to raise a flag as defined by the user for quality metrics
        """
        self._files = files
        self._genome_size_lower_bound = genome_size_lower_bound
        self._genome_size_upper_bound = genome_size_upper_bound
        self._minimum_N50_value = minimum_N50_value
        self._minimum_contig_length = minimum_contig_length
        self._unacceptable_num_contigs = unacceptable_num_contigs
        
    
    def create_quality_module_dataframe(self):
        """
        Goes through the files and creates a dataframe consisting of the file's genome length, N50 value and the number of contigs greater or equal to the minimum contig length as
        specified by the quality metrics. It also consists of the feedback for whether or not the file passed the quality metrics and if it didn't feedback on why it failed
        :return: A pd.dataframe containing the genome size, N50 value, number of contigs equal to or above our user defined minimum contig length
        as well as the results of our quality metrics (pass or fail) and the corresponding feedback
        """
        name_set=[]
        for myFile in self._files:
            name_set.append(path.splitext(path.basename(myFile))[0])

        files_contigs_and_genomes_lengths=self._get_files_contigs_and_genomes_lengths(self._files)

        files_genome_length_feedback = self._get_genome_length_feedback(files_contigs_and_genomes_lengths[1],self._genome_size_lower_bound,self._genome_size_upper_bound)
        
        files_N50_value_feedback=self._get_N50_feedback(files_contigs_and_genomes_lengths[0],files_contigs_and_genomes_lengths[1],self._minimum_N50_value)
        
        file_num_contigs_over_minimum_bp_feedback= self._get_num_contigs_over_minimum_bp_feedback(files_contigs_and_genomes_lengths[0],self._minimum_contig_length,self._unacceptable_num_contigs)
        
        quality_module = self._get_quality_module(files_genome_length_feedback,files_N50_value_feedback[1],file_num_contigs_over_minimum_bp_feedback[1])      
        quality_module_feedback = quality_module[0]
        quality_module_result = quality_module[1]

        #Module to represent our quality metric values, the index which is used to merge this module and the feedback module is the file names
        quality_metrics_module = pd.DataFrame([[file_name,genome_length,N50_value,num_contigs_over_minimum_bp] for file_name,genome_length,N50_value,num_contigs_over_minimum_bp in 
            zip(name_set,files_contigs_and_genomes_lengths[1],files_N50_value_feedback[0],file_num_contigs_over_minimum_bp_feedback[0])],
            columns=('Isolate ID','Genome Length','N50 value','Number of Contigs Greater Than Or Equal To '+str(self._minimum_contig_length)+' bp')).set_index('Isolate ID')

        #Module to represent the feedback for our quality metrics, the index which is used to merge this module and the quality metric value module is the file names 
        feedback_module = pd.DataFrame([[file_name,feedback,detailed_feedback] for file_name,feedback,detailed_feedback in zip(name_set,quality_module_result,quality_module_feedback)],
            columns=('Isolate ID','Quality Module','Quality Module Feedback')).set_index('Isolate ID')
        
        #Module to represent out quality metric values and their corresponding feedback
        quality_module_frame=quality_metrics_module.merge(feedback_module, on='Isolate ID', how='left')
        
        return quality_module_frame

    def _get_files_contigs_and_genomes_lengths(self,files):
        #This solution was taken almost directly from https://github.com/phac-nml/sistr_cmd and was in no way specifically designed for starAMR
        """
        Goes through the files and determines their genome length as well as the length of each contig
        :param files: The files for which we wish to determine the genome length as well as the length of each contig
        :return: An array where the first element is itself an array where each element represents the corresponding 
        file and is itself an array where each element is the length for the corresponding contig inside of this file.
        The second element is itself an array where each element is the genome length for the corresponding file
        """
        #This array will contain the contig lengths for all the files to be used for our quality metrics, in particular to determine if the files pass our N50
        #metric, and to determine if the files pass the number of contigs equal to or above the minimum contig length metric
        files_contig_lengths=[]
        #This array will contain the genome lengths for all the files to be used for our quality metrics, in particualr to determine if the files pass our genome 
        #length metric, and to calculate our N50 value 
        files_genome_lengths =[]
        for file in files:
            contig_lengths = [len(record.seq) for record in SeqIO.parse(file, 'fasta')]
            files_contig_lengths.append(contig_lengths)
            files_genome_lengths.append(sum(contig_lengths))

        return [files_contig_lengths, files_genome_lengths]

    def _get_genome_length_feedback(self,files_genome_lengths,lb_gsize,ub_gsize):
        """
        Goes through the files and determines whether or not they pass the quality metrics for genome length
        :param files_genome_lengths: An array where each element is the genome length of the corresponding file
        :param lb_gsize: The lower bound for the genome size as defined by the user for quality metrics
        :param ub_gsize: The upper bound for the genome size as defined by the user for quality metrics
        :return: An array where each element corresponds to the feedback (true or false) for the corresponding file in regards to the
        genome size quality metric
        """
        #The array contains the feedback of either True or false for whether or not the corresponding files pass the genome length quality metric, and
        #this feedback will be used to construc our quality module dataframe
        files_genome_feedback=[genome_length >= lb_gsize and genome_length <= ub_gsize for genome_length in files_genome_lengths]
        return files_genome_feedback

    def _get_N50_feedback(self,files_contigs_lengths,files_genome_lengths,minimum_N50):
        #For information on what N50 is and how it is calculated see https://en.wikipedia.org/wiki/N50,_L50,_and_related_statistics
        """
        Goes through the files and determines whether or not they pass the quality metrics for N50 value
        :param files_contigs_lengths: The lengths of the contigs for the files
        :param files_genome_lengths: An array where each element is the genome length of the corresponding file
        :param minimum_N50_value: The minimum N50 value as defined by the user for quality metrics
        :return: An array where the first element is itself an array where each element is the N50 value for 
        the corresponding file. The second element is itself an array where each element is the feedback (true or false) 
        for whether the corresponding file passes the N50 quality metrics
        """
        #This array is what we will return and will contain the files N50 values and their feedback for whether or not they pass the N50 quality metric
        #and will be used to construct our quality module dataframe
        feedback = []
        #This array will contain the N50 values for the files, it will be returned as the first element of feedback
        files_N50=[]
        #This array will contain the feedback of either True or false for whether or not the corresponding files pass the N50 quality metric, it will be
        #returned as the second element of feedback
        N50_feedback = []
        for file_genome_length,file_contigs_lengths in zip(files_genome_lengths,files_contigs_lengths):
            half_length=(file_genome_length)/2
            file_contigs_lengths.sort()
            contig_num=len(file_contigs_lengths)
            contig_index = 1
            sum_of_largest_contigs=0

            while contig_index < contig_num:
                if sum_of_largest_contigs+file_contigs_lengths[contig_num-contig_index] >=half_length:
                    break

                else: 
                    sum_of_largest_contigs=sum_of_largest_contigs+file_contigs_lengths[contig_num-contig_index]
                    contig_index=contig_index+1

            files_N50.append(file_contigs_lengths[contig_num-contig_index])


        for file_N50_value in files_N50:
            if file_N50_value > minimum_N50:
                N50_feedback.append(True)

            else:
                N50_feedback.append(False)


        feedback.append(files_N50)
        feedback.append(N50_feedback)

        return feedback
    
    def _get_num_contigs_over_minimum_bp_feedback(self,files_contigs_lengths,minimum_contig_length,unacceptable_num_contigs_over_minimum_bp):
        """
        Goes through the files and determines whether or not they pass the quality metrics for the acceptable number of contigs equal to or above the minimum contig length
        :param files_contigs_lengths: The lengths of the contigs for the files
        :param minimum_contig_length: The minimum contig length as defined by the user for quality metrics
        :param unacceptable_num_contigs: The number of contigs in a file, equal to or above our minimum contig length, for which to raise a flag as defined by the user for quality metrics
        :return: An array where the first element is itself an array where each element is the number of contigs equal to or above the minimum contig length for
        the corresponding file. The second element is itself an array where each element is the feedback (True or False) 
        for whether the corresponding file passes the acceptable number of contigs equal to or above the minimum contig length quality metric
        """
        #This array is what we will return and will contain for each of the files the number of contigs of length greater than or equal to the minimum contig length 
        #as well as the feedback for whether or not this number of contigs is greater than or equal to the unacceptable number of contigs and thus wether it passes or fails the quality metric
        #this array will be used to construct our quality module dataframe
        feedback=[]
        #This array will contain the number of contigs of length greater than or equal to the minimum contig length for the files, it will be returned as the first element of feedback
        file_num_contigs=[]
        #This array will contain the feedback of either True or False for whether or not the corresponding files pass the quality metrics for the acceptable number of contigs equal to 
        #or above the minimum contig length, it will be returned as the second element of feedback
        contigs_over_minimum_bp_feedback=[]
        for file_contigs_lengths in files_contigs_lengths:
            num_contigs = 0

            for contig in file_contigs_lengths:
                if contig >= minimum_contig_length:
                    num_contigs = num_contigs+1

            file_num_contigs.append(num_contigs)


        for file_num_contigs_over_minimum_bp in file_num_contigs:
            if file_num_contigs_over_minimum_bp >= unacceptable_num_contigs_over_minimum_bp:
                contigs_over_minimum_bp_feedback.append(False)

            else:
                contigs_over_minimum_bp_feedback.append(True)


        feedback.append(file_num_contigs)
        feedback.append(contigs_over_minimum_bp_feedback)

        return feedback

    def _get_quality_module(self,genome_length_feedback,N50_feedback,contigs_over_minimum_bp_feedback):
        """
        Goes through the files and for each provides detailed feedback for why they failed the quality metrics
        :param genome_length_feedback: An array where each element is the feedback (true or false) for the corresponding file in regards to the
        genome length quality metric
        :param N50_feedback: An array where each element is the feedback (true or false) for the corresponding file in regards to the
        the N50 quality metric
        :param contigs_over_minimum_bp_feedback: An array where each element is the feedback (true or false) for the corresponding file in regards to the
        the acceptable number of contigs equal to or above the minimum contig length quality metric
        :return: An array where the first element is itself an array where each element is the detailed quality metric feedback for
        the corresponding file. The second element is itself an array where each element is the feedback (true or false) 
        for whether the corresponding file passes all of the quality metrics
        """
        #This array is what we will return and will contain the files feedback for whether or not they pass the quality metrics as well as the specific metrics
        #the files fail on, if they fail any of the metrics and will be used to construct our quality module dataframe
        feedback = []
        #This array contains the feedback of either True or False for whether the corresponding file passes all the quality metrics 
        #it will be returned as the first element of feedback
        quality_parameter = []
        #This array contains the detailed feeback or either nothing if the corresponding file passes all the quality metrics or the metrics which the corresponding file
        #fails on and it will be returned as the second element of feedback
        quality_parameter_feedback = []
        for file_genome_length_feedback,file_N50_feedback,file_contigs_over_minimum_bp_feedback in zip(genome_length_feedback,N50_feedback,contigs_over_minimum_bp_feedback):
            if all([file_genome_length_feedback,file_N50_feedback,file_contigs_over_minimum_bp_feedback]):
                quality_parameter_feedback_for_file=""
                quality_parameter.append("Passed")

            else:
                failed_feedback=[]
                quality_parameter.append("Failed")
                if file_genome_length_feedback == False:
                    failed_feedback.append("Genome length is not within the acceptable length range [{},{}]".format(self._genome_size_lower_bound,self._genome_size_upper_bound))
                if file_N50_feedback == False:
                    failed_feedback.append("N50 value is not greater than the specified minimum value [{}]".format(self._minimum_N50_value))

                if file_contigs_over_minimum_bp_feedback == False:
                    failed_feedback.append("Number of Contigs with a length greater than or equal to the minimum Contig length [{}] exceeds the acceptable number [{}]".format(self._minimum_contig_length,self._unacceptable_num_contigs))

                quality_parameter_feedback_for_file = ' ; '.join(failed_feedback)

            quality_parameter_feedback.append(quality_parameter_feedback_for_file)


        feedback.append(quality_parameter_feedback)
        feedback.append(quality_parameter)

        return feedback




import logging
import unittest

import pandas as pd
from os import path

from staramr.results.QualityModule import QualityModule

logger = logging.getLogger('QualityModuleTest')

class QualityModuleTest(unittest.TestCase):

    def setUp(self):
        self.genome_size_lower_bound = 400
        self.genome_size_upper_bound = 600
        self.minimum_N50_value = 100
        self.minimum_contig_length = 300
        self.unacceptable_num_contigs = 10
        self.test_data_dir = path.join(path.dirname(__file__), '..', 'data')

    def testN50ExactlyMinimumValue(self):
        file = path.join(self.test_data_dir, "test-N50-Exactly-Minimum-Value.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-N50-Exactly-Minimum-Value', quality_module.index[0], 'File name not equal')
        self.assertEqual(100, quality_module['N50 value'].iloc[0], 'N50 vlaue not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; N50 value is not greater than the specified minimum value [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testN50OneBPLargerThanMinimumValue(self):
        file = path.join(self.test_data_dir, "test-N50-One-BP-Larger-Than-Minimum-Value.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-N50-One-BP-Larger-Than-Minimum-Value', quality_module.index[0], 'File name not equal')
        self.assertEqual(101, quality_module['N50 value'].iloc[0], 'N50 vlaue not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testN50SmallerThanMinimumValue(self):
        file = path.join(self.test_data_dir, "test-N50-Smaller-Than-Minimum-Value.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-N50-Smaller-Than-Minimum-Value', quality_module.index[0], 'File name not equal')
        self.assertEqual(20, quality_module['N50 value'].iloc[0], 'N50 vlaue not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; N50 value is not greater than the specified minimum value [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testN50MuchLargerThanMinimumValue(self):
        file = path.join(self.test_data_dir, "test-N50-Much-Larger-Than-Minimum-Value.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-N50-Much-Larger-Than-Minimum-Value', quality_module.index[0], 'File name not equal')
        self.assertEqual(1000, quality_module['N50 value'].iloc[0], 'N50 vlaue not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testN50Calculation(self):
        #tests to make sure N50 contig length +all contig lengths greater than it >= half of genome length, here we are testing the = part
        file = path.join(self.test_data_dir, "test-N50-Calculation.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-N50-Calculation', quality_module.index[0], 'File name not equal')
        self.assertEqual(102, quality_module['N50 value'].iloc[0], 'N50 vlaue not equal')
        self.assertEqual('Passed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('', quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testN50UnaffectedByEmptyContigs(self):
        file = path.join(self.test_data_dir, "test-N50-Unaffected-By-Empty-Contigs.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-N50-Unaffected-By-Empty-Contigs', quality_module.index[0], 'File name not equal')
        self.assertEqual(101, quality_module['N50 value'].iloc[0], 'N50 vlaue not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testGenomeSizeExactlyMinimum(self):
        file = path.join(self.test_data_dir, "test-Genome-Size-Exactly-Minimum.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-Genome-Size-Exactly-Minimum', quality_module.index[0], 'File name not equal')
        self.assertEqual(400, quality_module['Genome Length'].iloc[0], 'Genome length not equal')
        self.assertEqual('Passed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('', quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testGenomeSizeExactlyMaximum(self):
        file = path.join(self.test_data_dir, "test-Genome-Size-Exactly-Maximum.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-Genome-Size-Exactly-Maximum', quality_module.index[0], 'File name not equal')
        self.assertEqual(600, quality_module['Genome Length'].iloc[0], 'Genome length not equal')
        self.assertEqual('Passed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('', quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testGenomeSizeWithinAcceptedRange(self):
        file = path.join(self.test_data_dir, "test-Genome-Size-Within-Accepted-Range.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-Genome-Size-Within-Accepted-Range', quality_module.index[0], 'File name not equal')
        self.assertEqual(500, quality_module['Genome Length'].iloc[0], 'Genome length not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('N50 value is not greater than the specified minimum value [{}]'.format(self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testGenomeSizeSmallerThanMinimum(self):
        file = path.join(self.test_data_dir, "test-Genome-Size-Smaller-Than-Minimum.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-Genome-Size-Smaller-Than-Minimum', quality_module.index[0], 'File name not equal')
        self.assertEqual(200, quality_module['Genome Length'].iloc[0], 'Genome length not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; N50 value is not greater than the specified minimum value [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testGenomeSizeLargerThanMaximum(self):
        file = path.join(self.test_data_dir, "test-Genome-Size-Larger-Than-Maximum.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-Genome-Size-Larger-Than-Maximum', quality_module.index[0], 'File name not equal')
        self.assertEqual(700, quality_module['Genome Length'].iloc[0], 'Genome length not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; N50 value is not greater than the specified minimum value [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testGenomeSizeUnaffectedByEmptyContigs(self):
        file = path.join(self.test_data_dir, "test-Genome-Size-Unaffected-By-Empty-Contigs.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-Genome-Size-Unaffected-By-Empty-Contigs', quality_module.index[0], 'File name not equal')
        self.assertEqual(600, quality_module['Genome Length'].iloc[0], 'Genome length not equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('N50 value is not greater than the specified minimum value [{}]'.format(self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')
    
    #NCO=Number of contigs with length above minimum contig length	NCU=Number of contigs with length under minimum contig length	NC=Number of contigs whose length is exactly minimum contig length
    def testNCExactlyUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NC-Exactly-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NC-Exactly-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(10, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; Number of Contigs with a length greater than or equal to the minimum Contig length [{}] exceeds the acceptable number [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_contig_length,self.unacceptable_num_contigs), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCUExactlyUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NCU-Exactly-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NCU-Exactly-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(0, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCOneUnderUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NC-One-Under-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NC-One-Under-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(9, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCOOneUnderUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NCO-One-Under-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NCO-One-Under-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(9, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCUnacceptableByEmptyContigs(self):
        file = path.join(self.test_data_dir, "test-NC-Unacceptable-By-Empty-Contigs.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NC-Unacceptable-By-Empty-Contigs', quality_module.index[0], 'File name not equal')
        self.assertEqual(9, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCOMuchLowerThanUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NCO-Much-Lower-Than-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NCO-Much-Lower-Than-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(1, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Passed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('', quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCUMuchHigherThanUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NCU-Much-Higher-Than-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NCU-Much-Higher-Than-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(0, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; N50 value is not greater than the specified minimum value [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')

    def testNCOMuchHigherThanUnacceptable(self):
        file = path.join(self.test_data_dir, "test-NCO-Much-Higher-Than-Unacceptable.fasta")
        files = [file]
        quality_module = QualityModule(files,self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_N50_value,self.minimum_contig_length,self.unacceptable_num_contigs)
        quality_module=quality_module.create_quality_module_dataframe()
        self.assertEqual(1, len(quality_module.index), 'Invalid number of rows in results')
        self.assertEqual('test-NCO-Much-Higher-Than-Unacceptable', quality_module.index[0], 'File name not equal')
        self.assertEqual(110, quality_module['Number of Contigs Greater Than Or Equal To '+ str(self.minimum_contig_length) +' bp'].iloc[0], 'Number of Contigs Greater Than Or Equal To Our Minimum Contig Length Not Equal')
        self.assertEqual('Failed', quality_module['Quality Module'].iloc[0], 'Quality result not equal')
        self.assertEqual('Genome length is not within the acceptable length range [{},{}] ; Number of Contigs with a length greater than or equal to the minimum Contig length [{}] exceeds the acceptable number [{}]'.format(self.genome_size_lower_bound,self.genome_size_upper_bound,self.minimum_contig_length,self.unacceptable_num_contigs), quality_module['Quality Module Feedback'].iloc[0], 'Quality feedback not equal')


from staramr.blast.results.pointfinder.PointfinderHitHSP import PointfinderHitHSP
from staramr.blast.results.pointfinder.codon.CodonMutationPosition import CodonMutationPosition
from staramr.blast.results.pointfinder.nucleotide.NucleotideMutationPosition import NucleotideMutationPosition


class PointfinderHitHSPPromoter(PointfinderHitHSP):
    """
    This class represents a Pointfinder BLAST hit and is responsible for parsing the related BLAST hit.
    """

    def __init__(self, file, blast_record, database_name):
        """
        Creates a new PointfinderHitHSPPromoter from a promoter-related BLAST hit.
        :param file: The input file.
        :param blast_record: The Bio.Blast.Record this hit came from.
        :param database_name: The name of the BLAST database. This name must be named in the same way the
            Pointfinder promoters are named (ex: "embA_promoter_size_115bp").
        """

        self._parse_database_name(database_name)
        super().__init__(file, blast_record)

    def _get_mutation_positions(self, start):
        nucleotide_mutations = []
        codon_mutations = []

        amr_seq = self.get_amr_gene_seq()
        genome_seq = self.get_genome_contig_hsp_seq()

        amr_pos = 0

        # Get all the nucleotide match positions up to the offset:
        for i in range(self.offset):

            # Insertion: "-" in the reference:
            if amr_seq[i] == "-":
                # left side
                offset = i - amr_pos + self.offset + 1  # accounting for string index and reference index possibly being different
                mutation = NucleotideMutationPosition(i, amr_seq, genome_seq, start, offset=offset + 1)
                nucleotide_mutations.append(mutation)
            # Mismatch or Deletion:
            elif (amr_seq[i] != genome_seq[i]):
                offset = i - amr_pos + self.offset
                mutation = NucleotideMutationPosition(i, amr_seq, genome_seq, start, offset=offset + 1)
                nucleotide_mutations.append(mutation)
                amr_pos += 1
            # Match:
            else:
                amr_pos += 1

        # Merge adjacent nucleotide insertions:
        nucleotide_mutations_merged = []
        while(len(nucleotide_mutations) > 0):
            current = nucleotide_mutations.pop(0)

            # If the merged list has something in it
            # and the current mutation is an insertion
            # and the last item in the merged list is an insertion
            # and the last item in the merged list has the same position:
            if(len(nucleotide_mutations_merged) > 0
            and current.get_database_amr_gene_mutation() == 'ins'
            and nucleotide_mutations_merged[-1].get_database_amr_gene_mutation() == 'ins'
            and nucleotide_mutations_merged[-1].get_mutation_position() == current.get_mutation_position()):
                # Add the nucleotide to the existing:
                nucleotide_mutations_merged[-1]._database_amr_gene_mutation += current._database_amr_gene_mutation
                nucleotide_mutations_merged[-1]._input_genome_mutation += current._input_genome_mutation

            else:
                nucleotide_mutations_merged.append(current)

        # Get all the codon match positions after the offset:
        for i in range(self.offset, len(amr_seq)):

            # Insertion: "-" in the reference:
            if amr_seq[i] == "-":
                # left side
                offset = i - amr_pos + self.offset # accounting for string index and reference index possibly being different
                mutation = CodonMutationPosition(i - 1, amr_seq, genome_seq, start, offset=offset)
                codon_mutations.append(mutation)
            # Mismatch or Deletion:
            elif (amr_seq[i] != genome_seq[i]):
                offset = i - amr_pos + self.offset
                mutation = CodonMutationPosition(i, amr_seq, genome_seq, start, offset=offset)
                codon_mutations.append(mutation)
                amr_pos += 1
            # Match:
            else:
                amr_pos += 1

        codon_mutations_filtered = []
        codon_starts = []

        # Only return codon mutation position objects with unique codon start positions
        for m in codon_mutations:
            if m._codon_start not in codon_starts:
                codon_starts.append(m._codon_start)
                codon_mutations_filtered.append(m)

        # Combine lists and return all positions:
        combined_mutations = nucleotide_mutations_merged + codon_mutations_filtered

        return combined_mutations

    def _parse_database_name(self, database_name):
        """
        Parses the name of the database in order to obtain the promoter offset.
        The database name is expected to have the following format:

        [GENENAME]_promoter_size_[SIZE]bp

        example:

        embA_promoter_size_115bp
        """

        tokens = database_name.split("_")  # split the name into tokens
        size_string = tokens[len(tokens) - 1]  # get the last token
        size = int(size_string.replace('bp', ''))  # remove the 'bp' and convert to an int

        self.offset = size



import logging
from os import path
from typing import List

import pandas as pd

from staramr.blast.AbstractBlastDatabase import AbstractBlastDatabase

logger = logging.getLogger('PlasmidfinderBlastDatabase')

"""
A Class for pulling information from the PlasmidFinder database.
"""


class PlasmidfinderBlastDatabase(AbstractBlastDatabase):

    def __init__(self, database_dir: str, database_type: str = None) -> None:
        """
        Creates a new PlasmidfinderBlastDatabase.
        :param database_dir: The specific PlasmidFinder database directory.
        """

        super().__init__(database_dir)
        self.database_type = database_type
        self.database_names = None

        existing_database_types = PlasmidfinderBlastDatabase.get_database_types(database_dir)

        if database_type is not None:
            # gram positive is special since at one point it corresponded to a single fasta file
            # but later it corresponds to multiple fasta files
            if database_type == 'gram_positive' and 'gram_positive' not in existing_database_types:
                config_df = self.get_config_table(database_dir)
                self.database_names = config_df.loc[config_df["name"] == "Gram Positive", "db_prefix"].tolist()
            else:
                if (database_type == 'enterobacteriaceae') and (database_type not in existing_database_types):
                    logger.info(f'PlasmidFinder database type [{database_type}] is out of date, switching to [enterobacteriales]')
                    database_type = 'enterobacteriales'

                self.plasmidfinder_database_file = path.join(self.database_dir, database_type)
                self.plasmidfinder_database_file += self.fasta_suffix
    
                if (not path.isfile(self.plasmidfinder_database_file)):
                    raise Exception(
                        f"Error, could not find file [{self.plasmidfinder_database_file}]. "
                        f"plasmidfinder database type [{database_type}] is either incorrect or plasmidfinder database not installed properly")
                elif database_type not in existing_database_types:
                    raise Exception(f"Plasmidfinder database type [{database_type}] is not valid")
                else:
                    self.database_names = [database_type]

    def get_name(self) -> str:
        return 'plasmidfinder'

    def get_database_names(self) -> List[str]:
        if self.database_names is not None:
            names = self.database_names
        else:
            names = super().get_database_names()

        return names

    @classmethod
    def get_available_databases(cls) -> list:
        """
        A Class Method to get a list of plasmidfinder databases that are currently supported by staramr.
        :return: The list of database_type currently supported by staramr.
        """
        return ['gram_positive', 'enterobacteriaceae', 'enterobacteriales']

    @classmethod
    def get_config_table(cls, database_dir: str) -> pd.DataFrame:
        """
        A Class Method to get the config table from the plasmidfinder database root directory.
        :return: A DataFrame containing the config table.
        """
        config = pd.read_csv(path.join(database_dir, 'config'), sep='\t', comment='#', header=None,
                             names=['db_prefix', 'name', 'description'])
        return config

    @classmethod
    def get_database_types(cls, database_dir: str) -> list:
        """
        A Class Method to get the list of databases from the PlasmidFinder database root directory.
        :param database_dir: The PlasmidFinder database root directory.
        :return: A list of databases in Plasmidfinder.
        """
        config = cls.get_config_table(database_dir)

        return config['db_prefix'].tolist()



import math

from staramr.blast.results.pointfinder.codon.CodonMutationPosition import CodonMutationPosition

"""
A class defining a codon-based insertion mutation for PointFinder.

This class is needed to handle the special case of codon insertions needing to report
the codon to the LEFT of the insertion, not the sequence position of the insertion itself.
"""


class CodonInsertionPosition(CodonMutationPosition):

    def __init__(self, match_position, database_amr_gene_string, input_genome_blast_string, database_amr_gene_start, offset=0):
        """
        Creates a new CodonMutationPosition.
        :param match_position: The particular position (0-based index) of the BLAST match string for this mutation.
        :param database_amr_gene_string: The database amr gene string from BLAST.
        :param input_genome_blast_string: The genome BLAST string from the input genome.
        :param database_amr_gene_start: The start coordinates of the BLAST amr gene hit.
        :param offset: The amount to offset the mutation by (important for promoter mutations).
        """
        super().__init__(match_position, database_amr_gene_string, input_genome_blast_string, database_amr_gene_start, offset)

        # We need to correctly set the codon position to be 1 to the left (-1) of the observed insertion codon.
        self._codon_start = math.ceil(self._nucleotide_position_amr_gene / 3) - 1

    def __repr__(self):
        return (
            'CodonInsertionPosition(_database_amr_gene_start={_database_amr_gene_start}, _nucleotide_position_amr_gene={_nucleotide_position_amr_gene}, '
            '_codon_start={_codon_start}, _database_amr_gene_codon={_database_amr_gene_codon}, _input_genome_codon={_input_genome_codon})').format(
            **self.__dict__)

    def get_pointfinder_mutation_string(self):
        # Since the position codon insertions in the Pointfinder database is off by one, we need
        # to adjust the position in the report to show what Pointfinder shows.
        return self.get_database_amr_gene_mutation() + str(
            self.get_mutation_position() + 1) + self.get_input_genome_mutation()




import logging
from os import path

import pandas as pd
import Bio.Seq
from staramr.blast.results.pointfinder.codon.CodonMutationPosition import CodonMutationPosition
from staramr.blast.results.pointfinder.codon.CodonInsertionPosition import CodonInsertionPosition

from staramr.exceptions.GenotypePhenotypeMatchException import GenotypePhenotypeMatchException

"""
A Class storing information about the specific PointFinder database.
"""

logger = logging.getLogger('PointfinderDatabaseInfo')


class PointfinderDatabaseInfo:

    def __init__(self, database_info_dataframe, file=None):
        """
        Creates a new PointfinderDatabaseInfo.
        :param database_info_dataframe: A pd.DataFrame containing the information in PointFinder.
        :param file: The file where the pointfinder database info originates from.
        """
        self._pointfinder_info = database_info_dataframe
        self._file = file

        self._resistance_table_hacks(self._pointfinder_info)

    @classmethod
    def from_file(cls, file):
        """
        Builds a new PointfinderDatabaseInfo from the passed file containing PointFinder information on drug resistance
        mutations.
        :param file: The file containing drug resistance mutations.
        :return: A new PointfinderDatabaseInfo.
        """

        with open(file) as f:
            line = f.readline()
        
        line = line.lstrip("#")
        column_names = line.split()

        pointfinder_info = pd.read_csv(file, sep='\t', index_col=False, comment='#', header=None, names=column_names)

        return cls(pointfinder_info, file)

    @classmethod
    def from_pandas_table(cls, database_info_dataframe):
        """
        Builds a new PointfinderDatabaseInfo from the passed pd.DataFrame.
        :param database_info_dataframe: A pd.DataFrame containing the information in PointFinder.
        :return: A new PointfinderDatabaseInfo.
        """
        return cls(database_info_dataframe)
    
    @staticmethod
    def to_codons(regex_match):
        # Sometimes, the regex will match a string with a comma and return multiple matches.
        # Ex: TTCATGGAC,TTC
        # We need to make sure we handle these commas and multiple matches.
        entries = regex_match.string.split(",")

        for i in range(0, len(entries)):
            entries[i] = Bio.Seq.translate(entries[i], table='Standard').upper()

        return ",".join(entries)

    def _resistance_table_hacks(self, table):
        """
        A function implementing some hacks to try and fix mismatched strings in the pointfinder databases.
        These should be removed when the underlying database is corrected.
        :param table: The pointfinder resistance table to fix.
        :return: None, but modifies the passed table in place.
        """
        if self._file and 'salmonella' in str(self._file) and path.exists(
                path.join(path.dirname(self._file), '16S_rrsD.fsa')):
            logger.debug("Replacing [16S] with [16S_rrsD] for pointfinder organism [salmonella]")
            table[['Gene_ID']] = table[['Gene_ID']].replace('16S', '16S_rrsD')

        # We need to fix table entries where codon deletions are listed,
        # but no reference nucleotides are listed in the corresponding entry.
        # This will mean that this specific entry (for some reason)
        # is using 0-based nucleotide coordinates and nucleotides for the Res_codon
        # instead of 1-based codon coordinates and codons for the Res_codon.
        # However, we need make an exception for entries where "promoter", "16S", or "23S"
        # are found in the 'Gene_ID'.

        table.loc[(table['Ref_nuc'] == '-') & (table['Ref_codon'] == 'del')
        & (table['Gene_ID'].str.contains('promoter') == False)
        & (table['Gene_ID'].str.contains('16S') == False)
        & (table['Gene_ID'].str.contains('23S') == False), "Codon_pos"] += 3
        # We increment by 3 because the table is using base-0 coordinates
        # and we have base-1 coordindates for everything codon-related.
        # However, what's listed is nucleotide coordinates for the codon
        # mutation, so we actually need to bump it by 3 nucleotides.

        # Furthermore, we need to convert 'Res_codon' entries related to the above problem
        # (codon deletions) from nucleotides into codons. We choose to convert nucleotides into
        # codons instead of the other way around so that can more accurately compare observed mutations
        # to the table entries.
        table.loc[(table['Ref_nuc'] == '-') & (table['Ref_codon'] == 'del')
        & (table['Gene_ID'].str.contains('promoter') == False)
        & (table['Gene_ID'].str.contains('16S') == False)
        & (table['Gene_ID'].str.contains('23S') == False), "Res_codon"] = table.loc[(table['Ref_nuc'] == '-') & (table['Ref_codon'] == 'del')
        & (table['Gene_ID'].str.contains('promoter') == False)
        & (table['Gene_ID'].str.contains('16S') == False)
        & (table['Gene_ID'].str.contains('23S') == False), "Res_codon"].str.replace('[A-Z,]+', self.to_codons, regex=True)

    def _get_resistance_codon_match(self, gene, codon_mutation):
        table = self._pointfinder_info

        # We need to handle codon deletions as a special case:
        if(type(codon_mutation) is CodonMutationPosition
        and codon_mutation.get_input_genome_amino_acid() == 'del'):
            matches = table[(table['Gene_ID'] == gene)
                    # Codon deletions are listed using nucleotide coordinates
                    & (table['Codon_pos'] == codon_mutation.get_mutation_position() * 3)
                        # Such coords are denoted in nucleotide coordinates in the reference table for some reason,
                        # so we need to convert to nucleotide coordinates before making the comparison.
                    & (table['Ref_codon'] == codon_mutation.get_database_amr_gene_mutation())
                    & (table['Res_codon'].str.contains(codon_mutation.get_input_genome_mutation(), regex=False))]
        
        # We need to handle codon insertions as a special case:
        # Pointfinder mis-reports the position of codon insertions. For example:
        # ref:     ACG --- ACG
        # query:   ACG GGG ACG
        # ref_pos: 1       2
        # Pointfinder is incorrectly reporting the insertion as 2_3insG instead of the correct 1_2insG,
        # which is to say, the insertion happens between reference codon coordinates 1 and 2.
        # We need to shift by 1 in our interpretation.
        elif type(codon_mutation) is CodonInsertionPosition:
            matches = table[(table['Gene_ID'] == gene)
                    # Codon inerstions need to be shifted by 1:
                    & (table['Codon_pos'] == codon_mutation.get_mutation_position() + 1)
                    & (table['Ref_codon'] == codon_mutation.get_database_amr_gene_mutation())
                    & (table['Res_codon'].str.contains(codon_mutation.get_input_genome_mutation(), regex=False))]

        # Normal cases:
        else:
            matches = table[(table['Gene_ID'] == gene)
                            & (table['Codon_pos'] == codon_mutation.get_mutation_position())
                            & (table['Ref_codon'] == codon_mutation.get_database_amr_gene_mutation())
                            & (table['Res_codon'].str.contains(codon_mutation.get_input_genome_mutation(), regex=False))]

        if len(matches.index) > 1:
            # If more then one match, try to match Res_codon exactly to subselect
            matches_subset = matches[matches['Res_codon'] == codon_mutation.get_input_genome_mutation()]

            if len(matches_subset.index) >= 1:
                matches = matches_subset

        return matches

    def _get_resistance_nucleotide_match(self, gene, nucleotide_mutations):
        return self._get_resistance_codon_match(gene, nucleotide_mutations)

    def get_phenotype(self, gene, codon_mutation):
        """
        Gets the phenotype for a given gene and codon mutation from PointFinder.
        :param gene: The gene.
        :param codon_mutation: The codon mutation.
        :return: A string describing the phenotype.
        """
        match = self._get_resistance_codon_match(gene, codon_mutation)

        if len(match.index) > 0:
            return match['Resistance'].iloc[0]
        else:
            raise GenotypePhenotypeMatchException("Error, no match for gene=" + str(gene) + ", codon_mutation=" + str(codon_mutation))

    def get_resistance_codons(self, gene, codon_mutations):
        """
        Gets a list of resistance codons from the given gene and codon mutations.
        :param gene: The gene.
        :param codon_mutations: The codon mutations.
        :return: The resistance codons.
        """
        resistance_mutations = []

        for codon_mutation in codon_mutations:
            match = self._get_resistance_codon_match(gene, codon_mutation)
            if len(match.index) > 0:
                resistance_mutations.append(codon_mutation)

        return resistance_mutations

    def get_resistance_nucleotides(self, gene, nucleotide_mutations):
        """
        Gets a list of resistance nucleotides from the given gene and nucleotide mutations.
        :param gene: The gene.
        :param nucleotide_mutations: The nucleotide mutations.
        :return: The resistance nucleotides.
        """
        resistance_mutations = []

        for nucleotide_mutation in nucleotide_mutations:
            match = self._get_resistance_nucleotide_match(gene, nucleotide_mutation)
            if len(match.index) > 0:
                resistance_mutations.append(nucleotide_mutation)

        return resistance_mutations



"""
An Exception to be raised when there are issues matching a genotype to a phenotype.
"""


class GenotypePhenotypeMatchException(Exception):

    def __init__(self, msg):
        """
        Constructs a new GenotypePhenotypeMatchException
        :param msg: The Exception message.
        """
        super().__init__(msg)



import logging

from staramr.blast.results.AMRHitHSP import AMRHitHSP
from staramr.blast.results.pointfinder.codon.CodonMutationPosition import CodonMutationPosition
from staramr.blast.results.pointfinder.codon.CodonInsertionPosition import CodonInsertionPosition

import pandas as pd

logger = logging.getLogger('PointfinderHitHSP')

"""
Class storing a PointFinder BLAST hit/hsp.
"""


class PointfinderHitHSP(AMRHitHSP):

    def __init__(self, file, blast_record):
        """
        Creates a new PointfinderHitHSP.
        :param file: The input file.
        :param blast_record: The Bio.Blast.Record this hit came from.
        """
        super().__init__(file, blast_record)

    def get_amr_gene_name(self):
        """
        Gets the particular gene name for the PointFinder hit.
        :return: The gene name.
        """
        return self._blast_record['qseqid']

    def _get_mutation_positions(self, start):
        mutation_positions = []
        mutation_positions_filtered = []
        codon_starts = []

        amr_seq = self.get_amr_gene_seq()
        genome_seq = self.get_genome_contig_hsp_seq()

        amr_pos = 0

        for i in range(len(amr_seq)):

            # Insertion: "-" in the reference:
            if amr_seq[i] == "-":
                # left side
                offset = i - amr_pos  # accounting for string index and reference index possibly being different
                mutation = CodonInsertionPosition(i, amr_seq, genome_seq, start, offset=offset)
                mutation_positions.append(mutation)
            # Mismatch or Deletion:
            elif (amr_seq[i] != genome_seq[i]):
                offset = i - amr_pos
                mutation = CodonMutationPosition(i, amr_seq, genome_seq, start, offset=offset)
                mutation_positions.append(mutation)
                amr_pos += 1
            # Match:
            else:
                amr_pos += 1

        # Only return mutation position objects with unique codon start positions
        for m in mutation_positions:
            if m._codon_start not in codon_starts:
                codon_starts.append(m._codon_start)
                mutation_positions_filtered.append(m)

        # @formatter:off
        return mutation_positions_filtered
        # @formatter:on

    def get_mutations(self):
        """
        Gets a list of NucleotideMutationPosition for the individual mutations.
        :return: A list of NucleotideMutationPosition.
        """
        return self._get_mutation_positions(self.get_amr_gene_start())



import unittest

import pandas as pd

from staramr.blast.results.pointfinder.PointfinderHitHSP import PointfinderHitHSP
from staramr.exceptions.InvalidPositionException import InvalidPositionException


class PointfinderHitHSPTest(unittest.TestCase):

    def testBuildPointfinderHitHSPSuccess(self):
        blast_record = pd.Series({'sstart': 1, 'send': 10, 'qstart': 1, 'qend': 10, 'sstrand': 'plus'})

        PointfinderHitHSP(file=None, blast_record=blast_record)

    def testBuildPointfinderHitHSPFailInvalidSubjectCoords(self):
        blast_record = pd.Series({'sstart': 10, 'send': 1, 'qstart': 1, 'qend': 10, 'sstrand': 'plus'})

        self.assertRaises(InvalidPositionException, PointfinderHitHSP, None, blast_record)

    def testBuildPointfinderHitHSPInvalidQueryCoords(self):
        blast_record = pd.Series({'sstart': 1, 'send': 10, 'qstart': 10, 'qend': 1, 'sstrand': 'plus'})

        self.assertRaises(InvalidPositionException, PointfinderHitHSP, None, blast_record)

    def testGetMutations_MultipleIndels(self):
        # This tests correctly handling the following sequence of mutations:
        # insertion, insertion, deletion, deletion

        blast_record = pd.Series({
            "qseqid":"pmrA",
            "sseqid":"pmrA",
            "pident":98.222,
            "length":675,
            "qstart":1,
            "qend":669,
            "sstart":1,
            "send":669,
            "slen":669,
            "qlen":669,
            "sstrand":"plus",
            "sseq":"ATGAAAATTCTGATTGTTGAAGACGATCCCACGCTGTTATTGCAGGGACTGATTCTGGCGGCGCAAACCGAAGGCTACGCGTGCGATGGCGTGACAACCGCGCGGATGGCGGAACAAAGCCTTGAGGCAGGTCATTACAGCCTGGTGGTACTGGATTTAGGGTTACCCGACGAAGATGGACTGCATTTTCTCGCCCGTATCCGGCAGAAAAAATATACCCTGCCGGTACTGATCCTCACCAAAGCTCGCGATACGCTGACCGACAAAATCGCCGGGCTGGATGTCGGTGCCGACGACTATCTGGTGAAGCCTTTTGCGCTGGAAGAGTTACATGCCCGTATCCGCGCCCTGCTACGACGCCATAATAATCAGGGCGAAAGTGAGCTGATTGTTGGCAATCTGACGCTGAACATGGGTCGCCGTCAGGTATGGATGGGCGGTGAAGAGTTGATT---ACGCCCAAAGAATATGCTCTGCTGTCACGGTTAATGCTCAAAGCAGGCAGTCCGGTGCATCGGGAAATTCTCTACAACGACATCTATAACTGGGACAATGAACCCTCGACCAACACCCTGGAAGTGCATATCCACAAT---CGCGACAAAGTGGGCAAAGCCCGTATCCGCACCGTGCGCGGCTTTGGCTATATGCTGGTCGCGAATGAGGAAAACTAA",
            "qseq":"ATGAAAATTCTGATTGTTGAAGACGAT---ACGCTGTTATTGCAGGGACTGATTCTGGCGGCGCAAACCGAAGGCTACGCGTGCGATGGCGTGACAACCGCGCGGATGGCGGAACAAAGCCTTGAGGCAGGTCATTACAGCCTGGTGGTACTGGATTTAGGGTTACCCGACGAAGATGGACTGCATTTTCTCGCCCGTATCCGGCAGAAAAAATATACCCTGCCGGTACTGATCCTCACC---GCTCGCGATACGCTGACCGACAAAATCGCCGGGCTGGATGTCGGTGCCGACGACTATCTGGTGAAGCCTTTTGCGCTGGAAGAGTTACATGCCCGTATCCGCGCCCTGCTACGACGCCATAATAATCAGGGCGAAAGTGAGCTGATTGTTGGCAATCTGACGCTGAACATGGGTCGCCGTCAGGTATGGATGGGCGGTGAAGAGTTGATTCTGACGCCCAAAGAATATGCTCTGCTGTCACGGTTAATGCTCAAAGCAGGCAGTCCGGTGCATCGGGAAATTCTCTACAACGACATCTATAACTGGGACAATGAACCCTCGACCAACACCCTGGAAGTGCATATCCACAATCTGCGCGACAAAGTGGGCAAAGCCCGTATCCGCACCGTGCGCGGCTTTGGCTATATGCTGGTCGCGAATGAGGAAAACTAA",
            "plength":100.89686098654708})
        
        hit = PointfinderHitHSP("staramr/tests/unit/data/pmrA-multi-indel.fsa", blast_record)
        mutations = hit.get_mutations()

        self.assertEqual(mutations[0]._codon_start, 9, msg='Wrong codon position.')
        self.assertEqual(mutations[0]._database_amr_gene_codon, "---", msg='Wrong AMR codon.')
        self.assertEqual(mutations[0]._input_genome_codon, "CCC", msg='Wrong input codon.')

        self.assertEqual(mutations[1]._codon_start, 79, msg='Wrong codon position.')
        self.assertEqual(mutations[1]._database_amr_gene_codon, "---", msg='Wrong AMR codon.')
        self.assertEqual(mutations[1]._input_genome_codon, "AAA", msg='Wrong input codon.')

        self.assertEqual(mutations[2]._codon_start, 150, msg='Wrong codon position.')
        self.assertEqual(mutations[2]._database_amr_gene_codon, "CTG", msg='Wrong AMR codon.')
        self.assertEqual(mutations[2]._input_genome_codon, "---", msg='Wrong input codon.')

        self.assertEqual(mutations[3]._codon_start, 197, msg='Wrong codon position.')
        self.assertEqual(mutations[3]._database_amr_gene_codon, "CTG", msg='Wrong AMR codon.')
        self.assertEqual(mutations[3]._input_genome_codon, "---", msg='Wrong input codon.')
